"""
The main module of symmetric.
"""

import sys
import json
import bisect
import flask

import symmetric.logging
import symmetric.constants
import symmetric.endpoints
import symmetric.helpers
import symmetric.errors
import symmetric.openapi.utils
import symmetric.openapi.docs


class _SymmetricSingleton(type):

    """
    Singleton metaclass to limit the existance of the symmetric object to one.
    """

    def __call__(cls, *args, **kwargs):
        """Defines the class creation method."""
        if not hasattr(cls, "symmetric_instance"):  # Object does not exist
            # Create object
            cls.symmetric_instance = super().__call__(*args, **kwargs)
        return cls.symmetric_instance  # Return symmetric object


class _Symmetric(metaclass=_SymmetricSingleton):

    """
    Main class to encapsulate every important feature of the symmetric package.
    """

    # Define allowed HTTP methods
    __allowed_methods = [
        "GET",
        "PUT",
        "POST",
        "DELETE",
        "OPTIONS",
        "HEAD",
        "PATCH",
        "TRACE"
    ]

    def __init__(self):
        self.__app = flask.Flask(__name__)  # Create flask app object
        self.__endpoints = []
        self.__openapi_schema = None
        self.__server_token_name = symmetric.constants.API_SERVER_TOKEN_NAME
        self.__client_token_name = symmetric.constants.API_CLIENT_TOKEN_NAME
        self.setup()

    @property
    def endpoints(self):
        """Returns a list with the endpoints."""
        return self.__endpoints

    @property
    def openapi(self):
        """
        Returns the openapi schema. If it does not exist, it creates it
        and returns it.
        """
        if not self.__openapi_schema:
            self.__openapi_schema = symmetric.openapi.utils.get_openapi(
                self,
                symmetric.helpers.humanize(
                    symmetric.helpers.get_module_name(self)
                ) + " API"
            )
        return self.__openapi_schema

    @property
    def client_token_name(self):
        """Return the client token name."""
        return self.__client_token_name

    def setup(self):
        """Sets up the API."""
        # Set up the endpoint for the openapi json schema
        # pylint: disable=W0612
        @self.__app.route(symmetric.constants.OPENAPI_ROUTE)
        def openapi_schema():
            return self.openapi

        # Set up the endpoint for the interactive documentation
        # pylint: disable=W0612
        @self.__app.route(symmetric.constants.DOCUMENTATION_ROUTE)
        def docs():
            return symmetric.openapi.docs.get_redoc_html(
                symmetric.helpers.humanize(
                    symmetric.helpers.get_module_name(self)
                ) + " API"
            )

    def __call__(self, *args, **kwargs):
        """
        Enable WSGI servers to start with CLI utilities
        (like gunicorn module:symmetric).
        """
        return self.__app.__call__(*args, **kwargs)

    def set_client_token_name(self, client_token_name):
        """Changes the default client token name to :client_token_name."""
        if not isinstance(client_token_name, str):  # Manage wrong type
            error = ("Invalid client token name type "
                     "given on set_client_token_name call")
            raise symmetric.errors.InvalidTokenNameError(error)
        if not client_token_name:  # Manage empty client token name
            error = ("Empty client token name given "
                     "on set_client_token_name call")
            raise symmetric.errors.InvalidTokenNameError(error)
        # Set new client token name
        self.__client_token_name = client_token_name
        return True

    def set_server_token_name(self, server_token_name):
        """Changes the default server token name to :server_token_name."""
        if not isinstance(server_token_name, str):  # Manage wrong type
            error = ("Invalid server token name type "
                     "given on set_server_token_name call")
            raise symmetric.errors.InvalidTokenNameError(error)
        if not server_token_name:  # Manage empty server token name
            error = ("Empty server token name given "
                     "on set_server_token_name call")
            raise symmetric.errors.InvalidTokenNameError(error)
        # Set new server token name
        self.__server_token_name = server_token_name
        return True

    def router(self, route, methods=["post"], response_code=200,
               auth_token=False):
        """
        Decorator modifier. Recieves a route string, a list of HTTP methods, a
        response code and a boolean indicating whether or not to authenticate.
        The route gets format-checked. Returns the original function unchanged.
        """
        try:
            symmetric.helpers.parse_route(route)
        except symmetric.errors.IncorrectRouteFormatError as err:
            self.__app.logger.error(
                f"[[symmetric]] IncorrectRouteFormatError: {err}"
            )
            sys.exit(1)

        methods = [
            symmetric.helpers.verb(x) for x in methods
            if symmetric.helpers.verb(x) in _Symmetric.__allowed_methods
        ]

        def decorator(function):
            """
            Function decorator. Recieves the main function and wraps it as a
            flask endpoint. Returns the original unwrapped function.
            """
            # Decorate the wrapper
            @self.__app.route(
                route, methods=methods, endpoint=function.__name__
            )
            def wrapper(*args, **kwargs):
                """
                Function wrapper. The main function gets logged, the JSON body
                gets extracted from the request and gets unpacked as **kwargs
                to pass to the main function. Some precautions are also taken
                (namely a try/except combo). Returns the function's output
                jsonified with a response code.
                """
                try:
                    self.__log_request(flask.request, route, function)

                    # Get the body
                    body = flask.request.get_json()
                    request_headers = flask.request.headers
                    if not body:
                        body = {}

                    # Check for token authentication
                    symmetric.helpers.authenticate(
                        request_headers, auth_token, self.__client_token_name,
                        self.__server_token_name)

                    # Filter method parameters
                    parameters = symmetric.helpers.filter_params(
                        function, body, auth_token, self.__client_token_name)
                    return flask.jsonify(function(**parameters)), response_code
                except symmetric.errors.AuthenticationRequiredError as err:
                    # Error authenticating
                    self.__app.logger.error(
                        f"[[symmetric]] exception caught: {err}"
                    )
                    return flask.jsonify({}), 401
                except Exception as err:
                    self.__app.logger.error(
                        f"[[symmetric]] exception caught: {err}"
                    )
                    return flask.jsonify({}), 500

            # Save Endpoint
            try:
                self.__save_endpoint(
                    symmetric.endpoints.Endpoint(
                        route,
                        methods,
                        response_code,
                        function,  # Save unchanged function
                        wrapper,   # Save flask decorated function
                        auth_token
                    )
                )
            except symmetric.errors.DuplicatedRouteError as err:
                self.__app.logger.error(
                    f"[[symmetric]] DuplicatedRouteError: {err}"
                )
                sys.exit(1)

            return function  # Return unchanged function
        return decorator

    def run(self, *args, **kwargs):
        """Executes the main run function of the Flask object."""
        self.__app.run(*args, **kwargs)

    def generate_markdown_documentation(self, module_name):
        """
        Gets the documentation of every endpoint and assembles it into a
        markdown formatted string.
        """
        docs = f"# {symmetric.helpers.humanize(module_name)} "
        docs += "API Documentation\n\n"
        docs += ("Endpoints that require an authentication token should "
                 f"send it in a key named `{self.__client_token_name}` "
                 "inside the request headers.\n\n")
        raw_docs = [
            x.generate_markdown_documentation() for x in self.__endpoints]
        docs += "\n".join(raw_docs)
        return docs

    def __save_endpoint(self, endpoint):
        """Saves an endpoint object and sorts the endpoints list."""
        if endpoint in self.__endpoints:
            message = f"Endpoint '{endpoint.route}' was defined twice."
            raise symmetric.errors.DuplicatedRouteError(message)
        bisect.insort(self.__endpoints, endpoint)

    def __log_request(self, request, route, function):
        """
        Recieves a request object, a route string and a function and
        logs the request event. It also logs the json body of the request.
        """
        self.__app.logger.info(
            f"{request.method} request to '{route}' endpoint "
            f"('{function.__name__}' function)."
        )
        body = request.get_json()
        if body:
            self.__log_body(body)

    def __log_body(self, body):
        """Given a json request body, logs it."""
        self.__app.logger.info("Request Body:\n" + json.dumps(
            body,
            indent=2,
            sort_keys=False,
            ensure_ascii=False
        ))


# Create symmetric object
symmetric_object = _Symmetric()



"""
A module for every helper of the OpenAPI documentation generator.
"""

import symmetric.constants


def is_not_docs(route):
    """Checks if a route is not a documentation-related route."""
    # Check that the route is not the schema route
    not_schema = route != symmetric.constants.OPENAPI_ROUTE
    # Check that the route is not the interactive documentation route
    not_documentation = route != symmetric.constants.DOCUMENTATION_ROUTE

    # Return if the route is neither of the documentation routes
    return not_schema and not_documentation



"""
A module for documentation HTML methods.
"""

import flask

import symmetric.constants


def get_redoc_html(title):
    redoc_script = ("https://cdn.jsdelivr.net/npm/redoc/bundles/"
                    "redoc.standalone.js")
    google_fonts = ("https://fonts.googleapis.com/css?family=Montserrat:"
                    "300,400,700|Roboto:300,400,700")
    html = f"""
    <!DOCTYPE html>
    <html>
        <head>
            <title>{title}</title>
            <!-- needed for adaptive design -->
            <meta charset="utf-8"/>
            <meta
                name="viewport"
                content="width=device-width, initial-scale=1"
            >
            <link href="{google_fonts}" rel="stylesheet">
            <style>
            body {{
                margin: 0;
                padding: 0;
            }}
            </style>
        </head>
            <body>
            <redoc spec-url="{symmetric.constants.OPENAPI_ROUTE}">
            </redoc>
            <script src="{redoc_script}"></script>
        </body>
    </html>
    """
    return flask.render_template_string(html)



"""
Module to hold the openapi documentation constants.
"""


ANY_TYPE = [
    {
        "type": "string"
    },
    {
        "type": "number"
    },
    {
        "type": "integer"
    },
    {
        "type": "boolean"
    },
    {
        "type": "array"
    },
    {
        "type": "object"
    },
]



"""
Module to hold the openapi documentation creation utilities.
"""

import inspect
import functools

import symmetric.openapi.constants
import symmetric.openapi.helpers


def get_openapi_endpoint(endpoint):
    """Generate the OpenAPI documentation for :endpoint."""
    request_body = get_openapi_endpoint_body(endpoint)
    response_codes = get_openapi_endpoint_responses(endpoint)
    path_doc = {}
    for http_method in map(lambda x: x.lower(), endpoint.methods):
        path_doc[http_method] = {
            "description": endpoint.docstring,
            "responses": response_codes
        }
        has_props = bool(request_body["properties"])
        has_body = has_props or bool(request_body["additionalProperties"])
        if endpoint.has_token:
            path_doc[http_method]["security"] = [
                {
                    "APIKeyAuth": []
                }
            ]
        if has_body:
            path_doc[http_method]["requestBody"] = {
                "required": has_props,
                "content": {
                    "application/json": {
                        "schema": request_body
                    }
                }
            }
    return {
        endpoint.route: path_doc
    }


def get_openapi_endpoint_body(endpoint):
    """Assembles the JSON schema for the endpoint body."""
    params = inspect.getfullargspec(endpoint.function)
    schema = {}
    if params.args:
        parameters_amount = len(params.args)
        defaults_amount = 0 if not params.defaults else len(params.defaults)
        defaults_amount = len(params.defaults)
        no_defaults = parameters_amount - defaults_amount
        for ii in range(no_defaults):
            if params.args[ii] not in params.annotations:
                var_label = "oneOf"
                var_type = symmetric.openapi.constants.ANY_TYPE
            else:
                var_label = "type"
                var_type = symmetric.helpers.type_to_string(
                    params.annotations[params.args[ii]])
            schema[params.args[ii]] = {
                var_label: var_type
            }
        for jj in range(defaults_amount):
            if params.args[no_defaults + jj] not in params.annotations:
                var_label = "oneOf"
                var_type = symmetric.openapi.constants.ANY_TYPE
            else:
                var_label = "type"
                var_type = symmetric.helpers.type_to_string(
                    params.annotations[params.args[no_defaults + jj]])
            schema[params.args[no_defaults + jj]] = {
                var_label: var_type,
                "default": params.defaults[jj]
            }
    return {
        "type": "object",
        "properties": schema,
        "additionalProperties": params.varkw is not None
    }


def get_openapi_endpoint_responses(endpoint):
    """Gets the OpenAPI Schema error codes for the endpoint."""
    params = inspect.getfullargspec(endpoint.function)
    responses = {
        f"{endpoint.response_code}": {
            "$ref": "#/components/responses/SuccesfulOperation"
        },
        "500": {
            "$ref": "#/components/responses/InternalError"
        }
    }
    if endpoint.has_token:
        responses["401"] = {
            "$ref": "#/components/responses/UnauthorizedError"
        }
    if "return" in params.annotations:
        responses[f"{endpoint.response_code}"]["content"] = {
            "application/json": {
                "schema": {
                    "type": symmetric.helpers.type_to_string(
                        params.annotations["return"])
                }
            }
        }
    return responses


def get_openapi(sym_obj, title, version="0.0.1", openapi_version="3.0.3"):
    """
    Gets the OpenAPI spec of every endpoint and assembles it into a
    JSON formatted object.
    """
    return {
        "openapi": openapi_version,
        "info": {
            "title": title,
            "version": version
        },
        "paths": functools.reduce(
            lambda x, y: {**x, **y},
            [get_openapi_endpoint(endpoint) for endpoint in sym_obj.endpoints
                if symmetric.openapi.helpers.is_not_docs(endpoint.route)],
            {}
        ),
        "components": {
            "securitySchemes": {
                "APIKeyAuth": {
                    "type": "apiKey",
                    "in": "header",
                    "name": sym_obj.client_token_name
                }
            },
            "responses": {
                "SuccesfulOperation": {
                    "description": "Successful operation"
                },
                "UnauthorizedError": {
                    "description": "Invalid or non-existent authentication "
                                   "credentials."
                },
                "InternalError": {
                    "description": "Unexpected internal error (API method "
                                   "failed, probably due to a missuse of the "
                                   "underlying function)."
                }
            }
        }
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['symmetric', 'symmetric.cli', 'symmetric.openapi']

package_data = \
{'': ['*']}

install_requires = \
['flask>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['symmetric = symmetric.cli.core:dispatcher']}

setup_kwargs = {
    'name': 'symmetric',
    'version': '3.4.0',
    'description': 'A powerful yet lean wrapper over Flask to massively speed up API creations and enable super fast module-to-API transformations.',
    'long_description': '# Symmetric\n\n![PyPI - Version](https://img.shields.io/pypi/v/symmetric?style=for-the-badge&logo=python&color=306998&logoColor=%23fff&label=version)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/symmetric?style=for-the-badge&logo=python&color=306998&logoColor=%23fff)\n\nA powerful yet lean wrapper over **[Flask](https://github.com/pallets/flask)** to massively speed up **[API](https://en.wikipedia.org/wiki/Web_API)** creations and enable super fast module-to-**[API](https://en.wikipedia.org/wiki/Web_API)** transformations.\n\n![Tests Workflow](https://img.shields.io/github/workflow/status/daleal/symmetric/tests?label=tests&logo=github&style=for-the-badge)\n![Linters Workflow](https://img.shields.io/github/workflow/status/daleal/symmetric/linters?label=linters&logo=github&style=for-the-badge)\n\n## Why Symmetric?\n\nRaw developing speed and ease of use, that\'s why. While `Flask` is a powerful tool to have, getting it to work from scratch can be a bit of a pain, especially if you have never used it before. The idea behind `symmetric` is to be able to take any module **already written** and transform it into a working API in a matter of minutes, instead of having to design the module ground-up to work with `Flask` (it can also be used to build an API from scratch really fast). With `symmetric`, you will also get some neat features, namely:\n\n- Auto logging.\n- Server-side error detection and exception handling.\n- Native support for an authentication token on a per-endpoint basis.\n- Auto-generated `/docs` endpoint for your API with **interactive documentation**.\n- Auto-generated [OpenAPI Specification](https://swagger.io/docs/specification/about/) and Markdown documentation files for your API.\n\nThe [complete documentation](https://symmetric.one/docs/basic-usage/) is available on the [official website](https://symmetric.one/).\n\n## Installing\n\nInstall using pip!\n\n```bash\npip install symmetric\n```\n\n## Usage\n\n### Running the development server\n\nTo start the development server, just run:\n\n```bash\nsymmetric run <module>\n```\n\nWhere `<module>` is your module name (in the examples, we will be writing in a file named `module.py`, so the module name will be just `module`). A `Flask` instance will be spawned immediately and can be reached at [http://127.0.0.1:5000](http://127.0.0.1:5000) by default. We don\'t have any endpoints yet, so we\'ll add some later. **Do not use this in production**. The `Flask` server is meant for development only. Instead, you can use any `WSGI` server to run the API. For example, to run the API using [gunicorn](https://gunicorn.org/), you just need to run `gunicorn module:symmetric` and a production ready server will be spawned.\n\n### Defining the API endpoints\n\nThe module consists of a main object called `symmetric`, which includes an important element: the `router` decorator. Let\'s analyze it:\n\n```py\nfrom symmetric import symmetric\n\n@symmetric.router("/some-route", methods=["post"], response_code=200, auth_token=False)\n```\n\nThe decorator recieves 4 arguments: the `route` argument (the endpoint of the API to which the decorated function will map), the `methods` argument (a list of the methods accepted to connect to that endpoint, defaults in only `POST` requests), the `response_code` argument (the response code of the endpoint if everything goes according to the plan. Defaults to `200`) and the `auth_token` argument (a boolean stating if the endpoint requires authentication using a `symmetric` token. Defaults to `False`).\n\nNow let\'s imagine that we have the following method:\n\n```py\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nTo transform that method into an API endpoint, all you need to do is add one line:\n\n```py\n@symmetric.router("/sample", methods=["get"])\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nRun `symmetric run module` and send a `GET` request to `http://127.0.0.1:5000/sample`. You should get a `Hello World!` in response! (To try it with a browser, make sure to run the above command and click [this link](http://127.0.0.1:5000/sample)).\n\nBut what about methods with arguments? Of course they can be API\'d too! Let\'s now say that you have the following function:\n\n```py\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\nTo transform that method into an API endpoint, all you need to do, again, is add one line:\n\n```py\n@symmetric.router("/add")\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\n### Querying API endpoints\n\nTo give parameters to a function, all we need to do is send a `json` body with the names of the parameters as keys. Let\'s see how! Run `symmetric run module` and send a `POST` request (the default `HTTP` method) to `http://127.0.0.1:5000/add`, now using the `requests` module.\n\n```python\nimport requests\n\npayload = {\n    "a": 48,\n    "b": 21\n}\nresponse = requests.post("http://127.0.0.1:5000/add", json=payload)\nprint(response.json())\n```\n\nWe got a `69` response! (`48 + 21 = 69`). Of course, you can return dictionaries from your methods and those will get returned as a `json` body in the response object **automagically**!\n\nWith this in mind, you can transform any existing project into a usable API very quickly!\n\n## ReDoc Documentation\n\nBy default, you can `GET` the `/docs` endpoint (using a browser) to access to **interactive auto-generated documentation** about your API. It will include request bodies for each endpoint, response codes, authentication required, default values, and much more!\n\n**Tip**: Given that the [ReDoc Documentation](https://github.com/Redocly/redoc) is based on the OpenAPI standard, using **type annotations** in your code will result in a more detailed interactive documentation. Instead of the parameters being allowed to be any type, they will be forced into the type declared in your code. Cool, right?\n\n## Developing\n\nClone the repository:\n\n```bash\ngit clone https://github.com/daleal/symmetric.git\n\ncd symmetric\n```\n\nRecreate environment:\n\n```bash\n./environment.sh\n\n. .venv/bin/activate\n```\n\nTest install:\n\n```bash\npoetry install  # will also install the symmetric CLI\n```\n\nRun the tests:\n\n```bash\npython -m unittest\n```\n\n## Resources\n\n- [Official Website](https://symmetric.one/)\n- [Issue Tracker](https://github.com/daleal/symmetric/issues/)\n',
    'author': 'Daniel Leal',
    'author_email': 'dlleal@uc.cl',
    'maintainer': 'Daniel Leal',
    'maintainer_email': 'dlleal@uc.cl',
    'url': 'https://symmetric.one/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
